import React, { createContext, useContext, useEffect, useRef, useState } from 'react';

/**
 * Creates a Atomic store context with a provider and hooks to access the store
 * Atomic means that each value in the store updates state separately using useStoreValue hook
 * Extremely minimalistic implementation of Jotai's store context
 * @param initialState - the initial state of the store
 * @returns - {Provider, useStoreValue, useStoreSetState}
 */
function createStoreContext(initialState) {
  // returns unsubscribe function

  function useStoreData() {
    const storeRef = useRef(initialState);
    const getSnapshot = useRef(() => storeRef.current).current;
    const subscribersRef = useRef([]);
    const setState = useRef(partialStateOrFunc => {
      if (typeof partialStateOrFunc === 'function') {
        const value = partialStateOrFunc(storeRef.current);
        storeRef.current = {
          ...storeRef.current,
          ...value
        };
      } else {
        storeRef.current = {
          ...storeRef.current,
          ...partialStateOrFunc
        };
      }
      subscribersRef.current.forEach(callback => callback());
    }).current;
    const subscribe = useRef(callback => {
      subscribersRef.current.push(callback);
      return () => subscribersRef.current.filter(cb => cb !== callback);
    }).current;
    return {
      getSnapshot,
      setState,
      subscribe
    };
  }
  const StoreContext = /*#__PURE__*/createContext(null);
  function Provider(props) {
    const value = useStoreData();
    return /*#__PURE__*/React.createElement(StoreContext.Provider, {
      value: value
    }, props.children);
  }

  /**
   * @param selector
   * @returns
   *
   * @category Client State
   */
  function useStoreValue(selector) {
    const store = useContext(StoreContext);
    if (!store) {
      throw new Error('Store not found');
    }
    const [state, setState] = useState(selector(store.getSnapshot()));
    useEffect(() => store.subscribe(() => setState(selector(store.getSnapshot()))), [selector, store]);
    return state;
  }

  /**
   *
   * @returns
   *
   * @category Client State
   */
  function useStoreSetState() {
    const store = useContext(StoreContext);
    if (!store) {
      throw new Error('Store not found');
    }
    return store.setState;
  }
  return {
    Provider,
    useStoreValue,
    useStoreSetState
  };
}

/**
 * Exclude types from documentation site, but we should still add doc comments
 * @internal
 */

export const {
  Provider: StreamVideoStoreProvider,
  useStoreValue: useStreamVideoStoreValue,
  useStoreSetState: useStreamVideoStoreSetState
} = createStoreContext({
  videoDevices: [],
  audioDevices: [],
  currentVideoDevice: undefined,
  currentAudioDevice: undefined
});
//# sourceMappingURL=StreamVideoContext.js.map